class Headers(dict):
    def __init__(self, header: dict):
        super().__init__({key.title(): header[key] for key in header} if header else {})

    def __getitem__(self, key):
        return super().__getitem__(key.title())

    def __setitem__(self, key, value):
        super().__setitem__(key.title(), str(value))

    def __or__(self, other):
        t = super().__or__(other)
        for k in other:
            if other[k] is None:
                del t[k]
        return t

    def __ior__(self, other):
        t = super().__ior__(other)
        for k in other:
            if other[k] is None:
                del t[k]
        return t



import asyncio
from asyncio import Future, Semaphore, wait_for
from asyncio.exceptions import TimeoutError
from inspect import isawaitable, iscoroutine
from traceback import format_exc
from types import AsyncGeneratorType, GeneratorType

from nhmspider.core.interface import CrawlerABC
from nhmspider.core.scheduler import Scheduler
from nhmspider.exceptions import ClassTypeError, ExceptionEnum, NhmException, NoCrawlerError, StopEngine
from nhmspider.http.request import Request
from nhmspider.http.response import Response
from nhmspider.item.base import Item
from nhmspider.utils.log import get_logger
from nhmspider.utils.module_loading import import_string
from nhmspider.utils.request import request_fingerprint
from nhmspider.utils.signal import SignalManager


class Crawler(CrawlerABC):
    def __init__(self, spider_class):
        self.logger = get_logger("Crawler")
        self.spider = spider_class.from_crawler(crawler=self)

        downloader_class = import_string(self.spider.settings.get_string("DEFAULT_DOWNLOADER_CLASS"))
        base_downloader_class = import_string("nhmspider.core.interface.DownloaderABC")
        if not issubclass(downloader_class, base_downloader_class):
            raise ClassTypeError
        self.downloader = downloader_class(self.spider)
        self.scheduler = Scheduler(self.spider)

        self.concurrent_requests = self.spider.settings.get_integer("CONCURRENT_REQUESTS")

        # pipeline
        self.enabled_pipeline = []
        for enabled_pipeline_string in self.spider.settings.get_list("ENABLED_PIPELINE"):
            pipeline_class = import_string(enabled_pipeline_string)
            self.enabled_pipeline.append(pipeline_class())

        # download middleware
        self.enabled_download_middleware = []
        for middleware_class_string in self.spider.settings.get_list("ENABLED_DOWNLOAD_MIDDLEWARE"):
            middleware_class = import_string(middleware_class_string)
            self.enabled_download_middleware.append(middleware_class())

        # spider middleware
        # enabled_spider_middleware = settings.get_list("ENABLED_SPIDER_MIDDLEWARE")
        # self.enabled_spider_middleware = [cls() for cls in enabled_spider_middleware]
        self.__STOP_FLAG = None

    def init_status(self):
        self.__STOP_FLAG = False

    def stop(self):
        self.__STOP_FLAG = True

    async def run_forever(self):
        while 1:
            await self.run()
            await asyncio.sleep(self.spider.settings.get_integer("RUN_LOOP_INTERVAL"))

    async def _open_crawler(self):
        """
        初始化crawler
        """
        self.init_status()
        # todo: 应尝试减少某些模块的初始化次数
        await self.scheduler.open_scheduler()
        await self.downloader.open_downloader()
        await self.spider.custom_init()
        self.signal_manager = SignalManager(self.scheduler.request_queue)
        self.signal_manager.connect()

        # init pipeline
        for pipeline in self.enabled_pipeline:
            pipe = pipeline.open_spider(self.spider)
            if isawaitable(pipe):
                await pipe

        # init download middleware
        for middleware in self.enabled_download_middleware:
            middle = middleware.open_spider(self.spider)
            if isawaitable(middle):
                await middle

    async def _close_crawler(self):
        """
        退出crawler的准备操作
        """
        await self.downloader.close_downloader()

        # clear pipeline
        for pipeline in self.enabled_pipeline:
            pip = pipeline.close_spider(self.spider)
            if isawaitable(pip):
                await pip

        # clear download middleware
        for middleware in self.enabled_download_middleware:
            mid = middleware.close_spider(self.spider)
            if isawaitable(mid):
                await mid

    async def crawl(self):

        def callback(future: Future):
            semaphore.release()
            exception = future.exception()
            if exception:
                self.scheduler.request_queue.task_done()
                self.scheduler.request_count += 1
                raise exception

        # todo: 应打印初始化了哪些模块。

        await self._open_crawler()

        tasks = []
        semaphore = Semaphore(value=self.concurrent_requests)
        try:
            # 初始化
            results = self.spider.start_request()
            await self.process_results(results)
            # todo: heartbeat应放置到单独模块中去
            tasks.append(asyncio.create_task(self.scheduler.heartbeat()))
            while not self.__STOP_FLAG:
                # 强制退出时候退出循环 todo: 续判断退出到时候当前已经开始的任务是否已经执行完？
                if self.scheduler.request_queue._finished.is_set():
                    break
                # 所有任务都已经处理完时，执行退出循环
                if self.scheduler.request_queue.empty() and semaphore._value == self.concurrent_requests:
                    break
                try:
                    request = await wait_for(self.scheduler.next_request(), timeout=1)
                except TimeoutError:
                    continue
                await semaphore.acquire()
                asyncio.create_task(self.process(request)).add_done_callback(callback)

            if not self.__STOP_FLAG:
                # 阻塞并等待所有任务完成
                await self.scheduler.request_queue.join()
            else:
                self.stop()

            # 正常推出时执行的关闭
            success_close_task = self.spider.custom_success_close()
            if isawaitable(success_close_task):
                await success_close_task
        finally:
            await self._close_crawler()

            # 所有task完成后，取消任务，退出程序
            for task in tasks:
                task.cancel()
            # 等待task取消完成
            await asyncio.gather(*tasks, return_exceptions=True)

            spider_close_task = self.spider.custom_close()
            if isawaitable(spider_close_task):
                await spider_close_task

            # 清理内存，消除对 RUN_FOREVER = True 时的影响
            self.scheduler.dupe_memory_queue.clear()
            tasks.clear()

            # todo: 应打印采集完成汇总的数据。

    async def process_results(self, results, response=None):
        if results:
            if isinstance(results, GeneratorType):
                for obj in results:
                    await self.process_result_single(obj, response)
            elif isinstance(results, AsyncGeneratorType):
                async for obj in results:
                    await self.process_result_single(obj, response)
            elif isinstance(results, Request):
                await self.process_result_single(results, response)
            elif iscoroutine(results):
                await results
            else:
                # todo: 考虑如何处理
                self.logger.error(f"丢弃该任务，未处理的处理结果类型：{results}。")

    async def process_request(self, obj, response):
        """
        处理单个request，将request添加到请求队列中等待调度
        """

        # 处理request对象优先级，深度优先
        obj.priority = (response.request.priority - 1) if obj.priority is None and response is not None else 0
        # todo: 指纹存入request对象内，作为一个属性存在
        # 根据指纹去重
        fp = request_fingerprint(obj)
        if obj.dont_filter is True or fp not in self.scheduler.dupe_memory_queue:
            obj.fp = fp
            await self.scheduler.enqueue_request(obj)
            self.scheduler.dupe_memory_queue.add(fp)

    async def process_item(self, obj):
        """
        处理单个item，在pipeline中处理item
        """

        if not self.enabled_pipeline and self.spider.DEBUG is True:
            self.logger.info(obj)
        self.scheduler.item_count += 1
        for pipeline in self.enabled_pipeline:
            obj = pipeline.process_item(obj, self.spider)
            if isawaitable(obj):
                obj = await obj

    async def process_result_single(self, obj, response):
        """
        处理spider中返回的对象
        """
        if isinstance(obj, Request):
            await self.process_request(obj, response)
        elif isinstance(obj, Item):
            await self.process_item(obj)
        else:
            self.logger.warning(f"[yield]尚未处理的类型[{obj.__class__.__name__}]。")

    async def process(self, request: Request):
        response = await self.download_request(request)
        if isinstance(response, Response):
            if self.spider.DEBUG is True:
                self.logger.info(f"Crawled ({response.status}) {response}.")
        else:
            # todo: 待处理非response的情况
            # 失败的请求也要调用task_done，否则无法结束。
            self.scheduler.request_queue.task_done()
            self.scheduler.request_count += 1
            return

        # todo: process_spider_in
        # generator or async generator
        # 并不会实际执行

        # parse response
        results = request.callback(response)
        # todo: process_spider_out 非此位置
        try:
            await self.process_results(results, response)
        except Exception as exc:
            if isinstance(exc, StopEngine):
                self.stop()
            self.logger.error(format_exc())
        finally:
            self.scheduler.request_queue.task_done()
            self.scheduler.request_count += 1

    async def download_request(self, request):
        # process_request
        for middleware in self.enabled_download_middleware:
            result = middleware.process_request(request, self.spider)
            if isawaitable(result):
                result = await result

            if isinstance(result, Request):
                return await self.process_results(result)
            elif isinstance(result, Response):
                # 返回response则直接跳过process_request
                request = result
                break
            elif result is not None:
                self.logger.error(f"未知的对象类型，{result}。")
                raise TypeError(ExceptionEnum.TYPE_ERROR)

        # download
        if isinstance(request, Request):
            response = await self.downloader.send_request(request)
        elif isinstance(request, Response):
            response = request
        else:
            self.logger.error(f"未知的对象类型: {request}。")
            raise TypeError(ExceptionEnum.TYPE_ERROR)

        # process_response
        if isinstance(response, Response):
            for middleware in self.enabled_download_middleware:
                if response is None:
                    break
                result = middleware.process_response(request, response, self.spider)
                if isawaitable(result):
                    result = await result

                if isinstance(result, Request):
                    return await self.process_results(result)
                elif isinstance(result, Response):
                    response = result
                elif result is not None:
                    self.logger.error(f"未知的对象类型: {result}。")
                    raise TypeError("未知的对象类型")
                else:
                    response = None
        elif isinstance(response, Exception):
            response: Exception
            self.logger.debug("\n" + response.stack)
            self.logger.debug(response.message)
            for middleware in self.enabled_download_middleware:
                result = middleware.process_exception(request, response, self.spider)
                if isawaitable(result):
                    result = await result
                if isinstance(result, Request):
                    return await self.process_results(result)
                elif isinstance(result, Response):
                    response = result
                    break
                elif isinstance(result, NhmException):
                    raise result
                elif result is not None:
                    self.logger.error(f"未知的对象类型: {result}。")
                    raise TypeError("未知的对象类型")
            else:
                raise response
        else:
            self.logger.error(f"未知的对象类型，{response}。")
            raise TypeError(ExceptionEnum.TYPE_ERROR)

        return response


class CrawlerRunner:
    def __init__(self):
        self.crawlers = []

    def crawl(self, spider_class):
        crawler = Crawler(spider_class)
        self.crawlers.append(crawler)


class CrawlerProcess(CrawlerRunner):
    def start(self):
        """
        启动已添加的到爬虫列表中的爬虫。

        可同时运行一个或多个爬虫
        1. 运行单个爬虫时，会在当前进程里启动爬虫
        2. 运行多个爬虫时，会使用多进程，在每个进程里启动单独的爬虫。
        """

        if not self.crawlers:
            raise NoCrawlerError("use method `CrawlerProcess.crawl` add spider class.")
        elif len(self.crawlers) == 1:
            # 只有一个爬虫任务，在主进程中运行
            crawler: Crawler = self.crawlers[0]

            import logging.config

            logging.config.dictConfig(crawler.spider.settings.get_dict("LOGGING"))

            # 是否循环运行爬虫
            if crawler.spider.settings.get_bool("RUN_FOREVER"):
                asyncio.run(crawler.run_forever())
            else:
                asyncio.run(crawler.run())
        else:
            # todo: 使用多进程，每个进程运行单个爬虫
            pass



from typing import Optional

from aiohttp.typedefs import StrOrURL

from nhmspider.http.headers import Headers


class Request:
    def __init__(
        self,
        url=None,
        callback=None,
        errback=None,
        form=None,
        headers=None,
        cookies=None,
        method="get",
        meta=None,
        dont_filter=False,
        priority=None,
        proxy=None,
        timeout=None,
        *args,
        **kwargs,
    ):
        self.method: str = method
        self.url: str = url
        self.callback: callable = callback
        self.errback: callable = errback
        self.headers: Headers = Headers(headers)
        self.cookies: dict = cookies
        self.proxy: Optional[StrOrURL] = proxy
        self.form: dict = form
        self.meta: dict = meta or {}
        self.dont_filter: bool = dont_filter
        # 优先级，越小越先请求
        self.priority: int = priority if priority is None else -priority

        # 创建body，供指纹使用
        if method.lower() == "get":
            self.body = url.split("?", 1)[-1] if "?" in url else ""
        elif method.lower() == "post":
            self.body = "&".join(
                map(lambda _: "=".join(_), sorted(form.items(), key=lambda x: x[0])),
            )
        else:
            self.body = ""
        self.body = self.body.encode()
        self.timeout = timeout

    # 供优先级队列比较，比较的是Request.priority的大小

    def __lt__(self, other):
        return self.priority < other.priority

    def __gt__(self, other):
        return self.priority > other.priority

    def __str__(self):
        return f"<{self.method.upper()} {self.url}>"


class FormRequest(Request):
    def __init__(
        self,
        url=None,
        callback=None,
        errback=None,
        form=None,
        headers=None,
        cookies=None,
        meta=None,
        dont_filter=False,
        priority=None,
        proxy=None,
        *args,
        **kwargs,
    ):
        self.method = "post"
        super().__init__(
            url,
            callback,
            errback,
            form,
            headers,
            cookies,
            self.method,
            meta,
            dont_filter,
            priority,
            proxy,
            *args,
            **kwargs,
        )



import json

from parsel import Selector


class Response:
    def __init__(
        self,
        url=None,
        request=None,
        text=None,
        response=None,
        status=None,
        headers=None,
    ):
        self.url = url
        self.request = request
        self.text = text
        self.__response = response
        self.__selector = Selector(text=text)
        self.meta = request.meta
        self.status = status
        self.headers = headers

    def xpath(self, xpath_string):
        return self.__selector.xpath(xpath_string)

    def json(self):
        return json.loads(self.text)

    def __str__(self):
        return f"<{self.request.method.upper()} {self.url}>"



import asyncio
import time

from nhmspider.utils.log import get_logger

logger = get_logger("TimeCounter")


def time_limit(seconds: int | float = 0, display: bool = False, timeout: bool = None):
    """
    限制异步或同步函数的执行时长的装饰器，支持延迟显示。

    该装饰器可以设置最大运行时间和函数运行完成后的等待时间。
    如函数执行时间少于指定时长，装饰器会自动延迟以补足时间。
    支持异步函数和同步函数的使用。

    Args:
        seconds (int | float): 函数运行所需的最小时间。如果函数运行时间少于此值，
            将自动延迟补足，缺省值为0。
        display (bool): 是否显示函数的运行时间日志。缺省值为False。
        timeout (bool | None): 最大超时时间。若为None，则不做超时限制；
            若设置超时时长，则超出时将引发asyncio超时相关的错误。
            缺省值为None。

    Returns:
        Callable: 返回一个可用于装饰函数的装饰器。
    """

    def outer(func):
        async def wrap(*args, **kwargs):
            start_time = time.time()
            if asyncio.iscoroutinefunction(func):
                if timeout is None:
                    r = await func(*args, **kwargs)
                else:
                    r = await asyncio.wait_for(func(*args, **kwargs), timeout=timeout)
            else:
                if timeout is None:
                    r = await asyncio.to_thread(func, *args, **kwargs)
                else:
                    r = await asyncio.wait_for(asyncio.to_thread(func, *args, **kwargs), timeout=timeout)
            end_time = time.time()
            cost_time = end_time - start_time
            if display:
                logger.info(f"[{func.__name__}] cost time {cost_time}")
            if seconds and cost_time < seconds:
                await asyncio.sleep(seconds - cost_time)
            return r

        return wrap

    return outer



import asyncio

from nhmspider.core.interface import SpiderABC
from nhmspider.http import Request
from nhmspider.utils.log import get_logger
from nhmspider.utils.pqueue import SpiderPriorityQueue


class Scheduler:
    def __init__(self, spider: SpiderABC):
        self.logger = get_logger(self.__class__.__name__)
        self.request_queue = None
        self.signal_manager = None
        # 请求去重队列
        self.dupe_memory_queue = set()
        self.spider = spider
        self.item_count = 0
        self.request_count = 0
        self.__opened = False

    async def open_scheduler(self):
        self.request_queue = SpiderPriorityQueue()
        self.__opened = True

    def close_scheduler(self):
        self.__opened = False

    @property
    def is_opened(self):
        return self.__opened

    async def next_request(self):
        request = await self.request_queue.get()
        return request

    async def enqueue_request(self, request: Request):
        assert request.callback, "未指定回调方法。"
        await self.request_queue.put(request)

    async def heartbeat(self, heartbeat_interval: int = 60):
        """
        todo: 统计采集的页数，抓取的条数。
              考虑放到extensions中去
        """
        last_item_count = 0
        last_request_count = 0
        while True:
            request_speed = self.request_count - last_request_count
            last_request_count = self.request_count
            item_speed = self.item_count - last_item_count
            last_item_count = self.item_count
            self.logger.info(
                f"Crawled {last_request_count} pages (at {request_speed} pages/min), "
                f"scraped {last_item_count} items (at {item_speed} items/min), "
                f"queue size {self.request_queue.qsize()}.",
            )

            await asyncio.sleep(heartbeat_interval)



from nhmspider.core.interface import CrawlerABC, SpiderABC
from nhmspider.http.request import Request
from nhmspider.settings.settings_manager import SettingsManager
from nhmspider.utils.log import get_logger
from nhmspider.utils.project import get_default_settings


class Spider(SpiderABC):
    name = "Spider"
    start_urls = []
    custom_settings = {}

    _settings: SettingsManager
    crawler: CrawlerABC

    def __init__(self, *args, **kwargs):
        self.logger = get_logger(self.__class__.__name__)
        self.logger.info(f"{self.__class__.__name__} start.")

    @property
    def settings(self):
        return self._settings

    @classmethod
    def from_crawler(cls, crawler: CrawlerABC, *args, **kwargs):
        spider = cls(*args, **kwargs)
        spider._set_crawler(crawler)
        spider._set_spider(crawler)
        return spider

    def _set_crawler(self, crawler: CrawlerABC): ...

    def _set_spider(self, crawler):
        self.crawler = crawler
        # 获取 default_settings
        default_settings = get_default_settings()
        self._settings = SettingsManager(default_settings) | self.custom_settings
        self.DEBUG = self.settings.get_boolean("DEBUG")

    async def custom_init(self): ...

    async def custom_close(self): ...

    async def custom_success_close(self): ...

    def start_request(self):
        for url in self.start_urls:
            request = Request(url, callback=self.parse)
            yield request

    def parse(self, response): ...

    def __del__(self):
        self.logger.info(f"{self.__class__.__name__} closed.")



from typing import Union

from nhmspider.core.spider import Spider
from nhmspider.download_middleware.interface import DownloadMiddlewareAbc
from nhmspider.exceptions import NhmException
from nhmspider.http.request import Request
from nhmspider.http.response import Response


class DownloadMiddleware(DownloadMiddlewareAbc):
    def open_spider(self, spider: Spider) -> None:
        return None

    def process_request(
        self,
        request: Request,
        spider: Spider,
    ) -> Union[Request, Response, None]:
        return None

    def process_response(
        self,
        request: Request,
        response: Response,
        spider: Spider,
    ) -> Union[Request, Response, None]:
        return response

    def process_exception(
        self,
        request: Request,
        exception: NhmException,
        spider: Spider,
    ) -> Union[Request, Response, NhmException, None]:
        return None

    def close_spider(self, spider: Spider) -> None:
        return None



"""
copy form django==5.2.4
"""

import sys
from importlib import import_module


def cached_import(module_path, class_name):
    # Check whether module is loaded and fully initialized.
    if not (
        (module := sys.modules.get(module_path))
        and (spec := getattr(module, "__spec__", None))
        and getattr(spec, "_initializing", False) is False
    ):
        module = import_module(module_path)
    return getattr(module, class_name)


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    try:
        return cached_import(module_path, class_name)
    except AttributeError as err:
        raise ImportError('Module "{}" does not define a "{}" attribute/class'.format(module_path, class_name)) from err



from nhmspider import Item
from nhmspider.core.spider import Spider
from nhmspider.pipeline.interface import PipelineAbc


class Pipeline(PipelineAbc):
    def open_spider(self, spider: Spider) -> None:
        return

    def process_item(self, item: Item, spider: Spider) -> Item:
        return item

    def close_spider(self, spider: Spider) -> None:
        return



"""
copy from scrapy-2.5.0

"""

import hashlib
from typing import Dict, Optional, Tuple
from weakref import WeakKeyDictionary

from w3lib.url import canonicalize_url

from nhmspider.http import Request

_fingerprint_cache: "WeakKeyDictionary[Request, Dict[Tuple[Optional[Tuple[bytes, ...]], bool], str]]"
_fingerprint_cache = WeakKeyDictionary()


def to_bytes(text, encoding=None, errors="strict"):
    """Return the binary representation of ``text``. If ``text``
    is already a bytes object, return it as-is."""
    if isinstance(text, bytes):
        return text
    if not isinstance(text, str):
        raise TypeError(
            "to_bytes must receive a str or bytes " f"object, got {type(text).__name__}",
        )
    if encoding is None:
        encoding = "utf-8"
    return text.encode(encoding, errors)


def request_fingerprint(
    request: Request,
    keep_fragments: bool = False,
):
    headers: Optional[Tuple[bytes, ...]] = None
    cache = _fingerprint_cache.setdefault(request, {})
    cache_key = (headers, keep_fragments)
    if cache_key not in cache:
        fp = hashlib.sha1()
        fp.update(to_bytes(request.method))
        fp.update(
            to_bytes(canonicalize_url(request.url, keep_fragments=keep_fragments)),
        )
        fp.update(request.body or b"")
        cache[cache_key] = fp.hexdigest()
    return cache[cache_key]



from nhmspider.exceptions import SettingsTypeError


class SettingsManager(dict):
    """
    配置管理
    """

    def get_integer(self, key, default=None) -> int:
        try:
            return int(self.get(key, default or 0))
        except TypeError:
            raise SettingsTypeError(key, self[key], int)

    get_int = get_integer

    def get_float(self, key, default=None) -> float:
        try:
            return float(self.get(key, default or 0.0))
        except TypeError:
            raise SettingsTypeError(key, self[key], float)

    def get_dictionary(self, key, default=None) -> dict:
        try:
            return dict(self.get(key, default or {}))
        except TypeError:
            raise SettingsTypeError(key, self[key], dict)

    get_dict = get_dictionary

    def get_boolean(self, key, default=None) -> bool:
        try:
            return bool(self.get(key, default or False))
        except TypeError:
            raise SettingsTypeError(key, self[key], dict)

    get_bool = get_boolean

    def get_string(self, key, default=None) -> str:
        try:
            return str(self.get(key, default or ""))
        except TypeError:
            raise SettingsTypeError(key, self[key], dict)

    get_str = get_string

    def get_list(self, key, default=None) -> list:
        try:
            return list(self.get(key, default or []))
        except TypeError:
            raise SettingsTypeError(key, self[key], list)

    def __or__(self, other):
        return SettingsManager(super().__or__(other))

    def __ior__(self, other):
        return super().__ior__(other)

    def __str__(self):
        return f"<{self.__class__.__name__} {super().__str__()}>"

    def update(self, __m, **kwargs) -> None:
        return super().update(__m)



from typing import Literal, Optional

from playwright.async_api import Browser, async_playwright

from nhmspider.core.interface import DownloaderABC
from nhmspider.http import Request, Response


class PlaywrightDownloader(DownloaderABC):
    browser_type: Literal["chromium", "firefox", "webkit"]

    async def open_downloader(self):
        await super().open_downloader()
        self.browser_type = self.spider.settings.get_string("BROWSER_TYPE", "chromium")

    async def send_request(self, request: Request) -> Optional[Response | Exception]:
        if request.method.lower() != "get":
            return None

        playwright_kwargs = request.meta.get("playwright", {})
        screenshot = playwright_kwargs.get("screenshot")
        try:
            async with async_playwright() as playwright:
                launch_kwargs = {}
                if request.proxy:
                    launch_kwargs["proxy"] = {"server": request.proxy}
                async with await getattr(playwright, self.browser_type).launch(**launch_kwargs) as browser:
                    browser: Browser
                    page = await browser.new_page()
                    await page.goto(request.url, timeout=(request.timeout or self.timeout) * 1000)
                    text = await page.content()
                    response = Response(
                        url=page.url,
                        request=request,
                        text=text,
                        response=None,
                        status=200,
                        headers={},
                    )
                    if screenshot is not None:
                        response.meta["screenshot"] = await page.screenshot()
        except Exception as exception:
            return exception
        return response



from asyncio import PriorityQueue

from nhmspider.utils.log import get_logger


class SpiderPriorityQueue(PriorityQueue):
    def __init__(self, maxsize=0):
        self._finished = None
        self.logger = get_logger(self.__class__.__name__)
        super().__init__(maxsize=maxsize)

    def close(self):
        """
        封装私密方法，调用此方法会释放 asyncio.join() 的结果
        """
        self.logger.info(f"当前剩余请求{self.qsize()}。")
        self._finished.set()



from abc import ABC, abstractmethod
from logging import Logger
from typing import Any, AsyncGenerator, Generator, Optional

from nhmspider.http import Request, Response
from nhmspider.item.base import Item
from nhmspider.utils.log import get_logger
from nhmspider.utils.time_counter import time_limit


class CrawlerABC(ABC):
    spider: SpiderABC
    logger: Logger

    @time_limit(display=True)
    async def run(self):
        await self.crawl()

    @abstractmethod
    async def crawl(self):
        raise NotImplementedError


TypeSpiderOut = Generator[Request | Item, None, None] | AsyncGenerator[Request | Item, None]


class SpiderABC(ABC):
    # 爬虫类的名称，spider类的logger会使用此名称创建
    name: str
    # 启动爬虫的初始链接，该数组中的链接会以get方式发送请求。
    start_urls: list[str]
    # 当前spider的专用配置
    custom_settings: dict[str, Any]

    @abstractmethod
    def start_request(self) -> TypeSpiderOut: ...

    @abstractmethod
    def parse(self, response: Response) -> TypeSpiderOut: ...

    @classmethod
    @abstractmethod
    def from_crawler(cls, crawler: CrawlerABC, *args, **kwargs): ...


class DownloaderABC(ABC):
    def __init__(self, spider: SpiderABC):
        self.logger = get_logger(self.__class__.__name__)
        self.spider = spider
        self.headers = None
        self.timeout = self.spider.settings.get_float("REQUEST_TIMEOUT")
        self.__opened: bool = False

        self.clear_cookie = self.spider.settings.get_boolean("CLEAR_COOKIE")
        self.use_session = self.spider.settings.get_boolean("USE_SESSION")

    async def open_downloader(self):
        self.headers = self.spider.settings.get_dict("DEFAULT_REQUEST_HEADER")
        self.__opened = True

    async def close_downloader(self):
        self.__opened = False

    @property
    def is_opened(self) -> bool:
        return self.__opened

    @abstractmethod
    async def send_request(self, request: Request) -> Optional[Response | Exception]:
        raise NotImplementedError



from asyncio.exceptions import TimeoutError

from nhmspider.download_middleware.base import DownloadMiddleware
from nhmspider.utils.log import get_logger


class TimeoutDownloadMiddleware(DownloadMiddleware):
    def __init__(self):
        self.logger = get_logger(self.__class__.__name__)

        # 最大重试次数
        self.max_retry_times = 3

    def process_exception(self, request, exception, spider):
        if isinstance(exception, TimeoutError):
            request.dont_filter = True
            retry_times = request.meta.get("exception_retry_times", 1)
            if retry_times < self.max_retry_times:
                self.logger.info(
                    f"{request} exception: TimeoutError, retry {retry_times} time...",
                )
                request.meta["exception_retry_times"] = retry_times + 1
                return request
            else:
                self.logger.warning(
                    f"{request} exception error, retry {self.max_retry_times} times error.",
                )
                return None
        return None



from nhmspider import Request
from nhmspider.core.spider import Spider
from nhmspider.download_middleware.base import DownloadMiddleware
from nhmspider.http.response import Response
from nhmspider.utils.log import get_logger


class RetryDownloadMiddleware(DownloadMiddleware):
    ignore_http_error: dict
    success_http_code: list[int]

    def __init__(self):
        self.logger = get_logger(self.__class__.__name__)
        # 最大重试次数
        self.max_retry_times = 3

    def open_spider(self, spider: Spider):
        self.ignore_http_error = spider.settings.get_dict("IGNORE_HTTP_ERROR")
        self.success_http_code = spider.settings.get_list("SUCCESS_HTTP_CODE")

    def process_response(self, request: Request, response: Response, spider: Spider):
        if request.meta.get("dont_retry", False):
            return response

        if isinstance(response, Response) and response.status not in self.success_http_code:
            if response.status in self.ignore_http_error:
                # 处理忽略的错误状态码
                return response

            return self._retry(request, f"response status error: {response.status}")
        return response

    def process_exception(self, request: Request, exception: Exception, spider: Spider):
        # todo: 待设置指定异常重试
        # if (
        #     isinstance(exception, self.EXCEPTIONS_TO_RETRY)
        #     and not request.meta.get('dont_retry', False)
        # ):
        #     return self._retry(request, exception)
        return self._retry(request, f"{exception.__class__.__module__}.{exception.__class__.__name__}")

    def _retry(self, request: Request, reason: str):
        # todo: 待验证重试是否正确
        if request.dont_filter:
            request.dont_filter = True
        retry_times = request.meta.get("retry_times", 1)
        if retry_times < self.max_retry_times:
            self.logger.info(f"{request} {reason}, retry {retry_times} time...")
            request.meta["retry_times"] = retry_times + 1
            return request
        else:
            self.logger.warning(
                f"{request} retry max {self.max_retry_times} times error.",
            )
            return None



import logging

VERSION = "4.0"
# 是否使用session
USE_SESSION = True
# 是否清理session的cookie，USE_SESSION = False时不生效
CLEAR_COOKIE = False
# 并发数量，即启动的任务数量
CONCURRENT_REQUESTS = 8
# 默认请求头
DEFAULT_REQUEST_HEADER = {"User-Agent": f"nhm-spider/{VERSION}"}
# 是否开启调试日志
DEBUG = True
# 日志输出等级
DEBUG_LEVEL = "INFO"
# 默认请求超时时间，30秒
REQUEST_TIMEOUT = 30
# 默认开启的管道
ENABLED_PIPELINE = []
# 默认开启的中间件
ENABLED_DOWNLOAD_MIDDLEWARE = [
    "nhmspider.download_middleware.default_headers.DefaultRequestHeadersDownloadMiddleware",
    "nhmspider.download_middleware.retry.RetryDownloadMiddleware",
    "nhmspider.download_middleware.timeout.TimeoutDownloadMiddleware",
]
# 忽略的状态码错误
IGNORE_HTTP_ERROR = []
# 允许的成功的请求状态码
SUCCESS_HTTP_CODE = [200]
# 是否循环执行爬虫
RUN_FOREVER = False
# 每次采集完等待间隔开始下一轮
# 默认：1天
RUN_LOOP_INTERVAL = 60 * 60 * 24
# 默认下载器
DEFAULT_DOWNLOADER_CLASS = "nhmspider.core.downloader.AiohttpDownloader"

LOG_LEVEL = logging.getLevelName(logging.DEBUG)
LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "verbose": {
            "format": "[%(levelname)s] %(asctime)s.%(msecs).3d %(filename)s(%(lineno)s) > %(funcName)s: %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
        },
        "simple": {"format": "[%(levelname)s] %(asctime)s.%(msecs).3d: %(message)s", "datefmt": "%Y-%m-%d %H:%M:%S"},
        "console": {
            "format": "[%(levelname)s] %(asctime)s.%(msecs).3d %(name)s: %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "console",
        },
    },
    "root": {
        "handlers": ["console"],
        "level": LOG_LEVEL,
    },
}



from abc import ABC, abstractmethod
from typing import Any

from nhmspider.exceptions import ValidationError


class BaseField(ABC):
    def __init__(self):
        self.__value = None

    def __set__(self, instance, value):
        self.validate(value)
        self.__value = value

    def __get__(self, instance, owner):
        return self.__value

    @abstractmethod
    def validate(self, value: Any) -> bool:
        raise NotImplemented


class Field(BaseField):
    def validate(self, value: Any) -> bool: ...


class IntegerField(Field):
    def validate(self, value: Any):
        if not isinstance(value, int):
            raise ValidationError(f"The value of an {self.__class__.__name__} instance must be an integer.")


class StringField(Field):
    def validate(self, value: Any):
        if not isinstance(value, str):
            raise ValidationError(f"The value of an {self.__class__.__name__} instance must be a string.")


class FloatField(Field):
    def validate(self, value: Any):
        if not isinstance(value, float):
            raise ValidationError(f"The value of an {self.__class__.__name__} instance must be a float.")


class BooleanField(Field):
    def validate(self, value: Any):
        if not isinstance(value, bool):
            raise ValidationError(f"The value of an {self.__class__.__name__} instance must be a boolean.")


class ListField(Field):
    def validate(self, value: Any):
        if not isinstance(value, list):
            raise ValidationError(f"The value of an {self.__class__.__name__} instance must be a list.")


class DictField(Field):
    def validate(self, value: Any):
        if not isinstance(value, dict):
            raise ValidationError(f"The value of an {self.__class__.__name__} instance must be a dict.")


class JsonField(Field):
    def validate(self, value: Any):
        import json

        try:
            json.dumps(value)
        except (TypeError, ValueError):
            raise ValidationError(f"The value of an {self.__class__.__name__} instance must be JSON serializable.")



# tests/test_time_counter.py

import asyncio
import time

from src.nhmspider.utils.time_counter import time_limit


def test_time_limit_synchronous_delay():
    @time_limit(seconds=2, display=False)
    def example_function():
        return "finished"

    start_time = time.time()
    result = asyncio.run(example_function())
    end_time = time.time()

    assert result == "finished"
    assert end_time - start_time >= 2  # Ensure delay occurs


def test_time_limit_asynchronous_delay():
    @time_limit(seconds=3, display=False)
    async def example_coroutine():
        return "completed"

    start_time = time.time()
    result = asyncio.run(example_coroutine())
    end_time = time.time()

    assert result == "completed"
    assert end_time - start_time >= 3  # Ensure delay occurs


def test_time_limit_no_delay():
    @time_limit(seconds=0.5, display=False)
    def quick_function():
        return "done"

    start_time = time.time()
    result = asyncio.run(quick_function())
    end_time = time.time()

    assert result == "done"
    assert end_time - start_time < 1  # Verify no unnecessary delays


def test_time_limit_timeout():
    @time_limit(seconds=0, display=False, timeout=2)
    async def slow_coroutine():
        await asyncio.sleep(5)
        return "error"

    try:
        asyncio.run(slow_coroutine())
    except asyncio.TimeoutError:
        assert True  # Ensure timeout exception is raised
    else:
        assert False, "Expected TimeoutError was not raised"


def test_time_limit_with_logging_behavior(caplog):
    @time_limit(seconds=1, display=True)
    async def example_log_function():
        return "logged"

    with caplog.at_level("INFO"):
        asyncio.run(example_log_function())

    log_messages = [record.message for record in caplog.records]
    assert any("cost time" in message for message in log_messages)



from nhmspider.core.spider import Spider
from nhmspider.crawler import CrawlerProcess
from nhmspider.http import FormRequest, Request, Response
from nhmspider.item.base import Item
from nhmspider.item.field.base import Field



from nhmspider.settings import default_settings
from nhmspider.settings.settings_manager import SettingsManager


def get_default_settings():
    settings = {}
    for key in dir(default_settings):
        if key.isupper():
            settings[key] = getattr(default_settings, key)
    return SettingsManager(settings)



from abc import ABC, abstractmethod
from typing import Union

from nhmspider.core.spider import Spider
from nhmspider.exceptions import NhmException
from nhmspider.http.request import Request
from nhmspider.http.response import Response


class DownloadMiddlewareAbc(ABC):
    @abstractmethod
    def open_spider(self, spider: Spider) -> None:
        """
        启动爬虫时，初始化中间件的方法

        :param spider: spider对象
        :type spider: scrapy.spider.base.SpiderAbc
        :return: 无返回值
        :rtype: None
        """

    @abstractmethod
    def process_request(
        self,
        request: Request,
        spider: Spider,
    ) -> Union[Request, Response, None]:
        """
        请求发送前的拦截器，可以在此添加对request的处理

        :param request: request对象
        :type request: scrapy.http.request.Request
        :param spider: spider对象
        :type spider: scrapy.spider.base.SpiderAbc
        :return: 请求 或 响应 对象。
                1. 当返回none时，会继续进行之后的中间件处理
                2. 当返回response时，会直接返回当前response对象，不会继续发送请求。
                3. 当返回request时，会将此请求重新添加到request队列中，重新进行调度。
                默认返回：none。
        :rtype: Union[Request, Response, None]
        """

    @abstractmethod
    def process_response(
        self,
        request: Request,
        response: Response,
        spider: Spider,
    ) -> Union[Request, Response, None]:
        """
        请求返回后的拦截器，可以在此添加对response的处理

        :param request: request对象
        :type request: scrapy.http.request.Request
        :param response: response对象
        :type response: scrapy.http.response.Response
        :param spider: spider对象
        :type spider: scrapy.spider.base.SpiderAbc
        :return: 请求 或 响应 对象。
                1. 当返回none时，会丢弃当前request。
                2. 当返回response时，会继续进行之后的中间件处理。
                3. 当返回request时，会将此请求重新添加到request队列中，重新进行调度。
                默认返回：response。
        :rtype: Union[Request, Response, None]
        """

    @abstractmethod
    def process_exception(
        self,
        request: Request,
        exception: NhmException,
        spider: Spider,
    ) -> Union[Request, Response, NhmException, None]:
        """
        下载异常的拦截器，可以在此添加对exception的处理

        :param request: request对象
        :type request: scrapy.http.request.Request
        :param exception: exception对象
        :type exception: scrapy.exceptions.NhmException
        :param spider: spider对象
        :type spider: scrapy.spider.base.SpiderAbc
        :return: 请求 或 响应 对象。
                1. 当返回none时，会丢弃request。
                2. 当返回response时，会跳过异常处理，直接返回当前response。
                3. 当返回request时，会将此请求重新添加到request队列中，重新进行调度。
                4. 当返回exception时，会继续进行之后中间件的处理。
                默认返回：response。
        :rtype: Union[Request, Response, None]
        """

    @abstractmethod
    def close_spider(self, spider: Spider) -> None:
        """
        关闭爬虫时，退出中间件的方法

        :param spider: spider对象
        :type spider: scrapy.spider.base.SpiderAbc
        :return: 无返回值
        :rtype: None
        """



from typing import Optional

import aiohttp
from aiohttp import ClientSession, ClientTimeout
from aiohttp_socks import ProxyConnector

from nhmspider.core.interface import DownloaderABC, SpiderABC
from nhmspider.http import Request, Response


class AiohttpDownloader(DownloaderABC):
    def __init__(self, spider: SpiderABC):
        super().__init__(spider=spider)
        self.sessions = {}

    async def get_session(self, request=None):
        proxy = request.proxy if request and request.proxy else None
        if (session := self.sessions.get(proxy)) is not None:
            return session

        async def on_request_start(session, trace_config_ctx, params):
            # print("Starting request")
            pass

        async def on_request_end(session, trace_config_ctx, params):
            # print("Ending request")
            pass

        trace_config = aiohttp.TraceConfig()
        trace_config.on_request_start.append(on_request_start)
        trace_config.on_request_end.append(on_request_end)

        connector = ProxyConnector.from_url(request.proxy) if proxy else None
        self.sessions[proxy] = aiohttp.ClientSession(
            connector=connector,
            headers=self.headers,
            timeout=ClientTimeout(self.timeout),
            trace_configs=[trace_config],
        )
        return await self.get_session(request)

    async def remove_session(self, request):
        proxy = request.proxy if request and request.proxy else None
        session = self.sessions.pop(proxy, None)
        if session is not None:
            await session.close()

    async def close_downloader(self):
        await super().close_downloader()
        for key in tuple(self.sessions.keys()):
            session = self.sessions.pop(key)
            if isinstance(session, ClientSession):
                await session.close()

    async def send_request(self, request) -> Optional[Response | Exception]:
        try:
            # 是否每次创建新session请求
            if self.use_session is False:
                session = await self.get_session(request)
                response = await self.send(session, request)
                await self.remove_session(request)
            else:
                session = await self.get_session(request)
                # 每次请求前清除session缓存的cookies 为response set-cookie中自动缓存的
                if self.clear_cookie is True:
                    session.cookie_jar.clear()
                response = await self.send(session, request)
            if response is None:
                return
            # 获取完text之后，会自动关闭response。
            text = await response.text()  # TimeoutError
        except Exception as exception:
            return exception
        return Response(
            request.url,
            request,
            text,
            response,
            response.status,
            response.headers,
        )

    async def send(self, session: aiohttp.ClientSession, request: Request) -> Optional[Response]:
        """处理不同method的请求参数"""
        timeout = aiohttp.ClientTimeout(total=(request.timeout or self.timeout))
        if request.method.lower() == "get":
            response = await session.get(
                request.url,
                data=request.body,
                headers=request.headers,
                cookies=request.cookies,
                timeout=timeout,
            )
        elif request.method.lower() == "post":
            response = await session.post(
                request.url,
                data=request.form,
                headers=request.headers,
                cookies=request.cookies,
                timeout=timeout,
            )
        else:
            response = None
        return response



from abc import ABC, abstractmethod

from nhmspider.core.spider import Spider
from nhmspider.item.base import Item


class PipelineAbc(ABC):
    @abstractmethod
    def open_spider(self, spider: Spider) -> None:
        """
        启动爬虫时，初始化管道的方法

        :param spider: spider对象
        :type spider: scrapy.spider.base.SpiderAbc
        :return: 无返回值
        :rtype: None
        """

    @abstractmethod
    def process_item(self, item: Item, spider: Spider) -> Item:
        """
        爬虫返回的item数据处理方法。

        :param item: spider返回的item，采集到的数据的载体。
        :type item: nhmspider.item.Item
        :param spider: spider对象
        :type spider: scrapy.spider.base.SpiderAbc
        :return: 返回数据的item对象，
        :rtype: nhmspider.item.Item
        """

    @abstractmethod
    def close_spider(self, spider: Spider) -> None:
        """
        关闭爬虫时，退出管道的方法

        :param spider: spider对象
        :type spider: scrapy.spider.base.SpiderAbc
        :return: 无返回值
        :rtype: None
        """



from pprint import pformat

from nhmspider.item.field.base import Field


class ItemMeta(type):
    def __new__(cls, name, bases, attrs):
        instance = super().__new__(cls, name, bases, attrs)
        meta = set()
        for key in attrs:
            if isinstance(attrs[key], Field):
                meta.add(key)
        instance.__meta = frozenset(meta)
        return instance

    @property
    def meta(cls):
        return cls.__meta


class BaseItem(metaclass=ItemMeta):
    def __init__(self, fields: dict = None, **kwargs):
        self._meta = self.__class__.meta
        if fields:
            for field in fields:
                if field not in self._meta:
                    raise AttributeError(
                        f"Class {self.__class__.__name__} not exists field [{field}].",
                    )
                setattr(self, field, fields[field])

        for field in kwargs:
            if field not in self._meta:
                raise AttributeError(
                    f"Class {self.__class__.__name__} not exists field [{field}].",
                )
            setattr(self, field, kwargs[field])

    def __repr__(self):
        return f"{self.__class__.__name__}({pformat(self.to_dict())})"

    def to_dict(self):
        return {key: getattr(self, key) for key in self._meta}


class Item(BaseItem):
    pass



from enum import Enum


class ExceptionEnum(Enum):
    """
    异常的错误信息提示
    """

    TYPE_ERROR = "未知的对象类型"


class NhmException(Exception):
    pass


class StopEngine(NhmException):
    """
    主动抛出以停止采集
    """


class SettingsError(NhmException):
    pass


class SettingsTypeError(SettingsError):
    """
    获取设置时转换类型错误
    """

    def __init__(
        self,
        key,
        value,
        _type,
    ):
        self.__key = key
        self.__value = value
        self.__type = _type

    def __str__(self):
        return f"Settings keyword argument [{self.__key}] need type {self.__type} but value is [{self.__value}]."


class NoCrawlerError(NhmException):
    """
    没有添加的爬虫任务
    """


class ClassTypeError(NhmException):
    def __str__(self):
        return "未继承基类的class"


class ValidationError(NhmException):
    pass
